import React, { useReducer, useRef, useState, useEffect } from 'react';
import { lowlight } from 'lowlight/lib/core';
import { toHtml } from 'hast-util-to-html';

var OPENING_BRACKETS = ['{', '[', '('];
var CLOSING_BRACKETS = ['}', ']', ')'];
var MUTLI_LINE_QUOTE = '`';
var SINGLE_LINE_QUOTES = ['"', "'"];
var SPECIAL_CHARACTERS = /*#__PURE__*/[].concat(OPENING_BRACKETS, CLOSING_BRACKETS, SINGLE_LINE_QUOTES, [MUTLI_LINE_QUOTE]);

var getCaretPosition = function getCaretPosition() {
  var selection = window.getSelection();
  var start = 0;
  var end = 0;

  if (selection) {
    var _ref = [selection.anchorOffset, selection.focusOffset];
    start = _ref[0];
    end = _ref[1];

    if (start > end) {
      var _ref2 = [end, start];
      start = _ref2[0];
      end = _ref2[1];
    }
  }

  return {
    start: start,
    end: end
  };
};

var restoreCaretPosition = function restoreCaretPosition(_ref3, node) {
  var start = _ref3.start,
      end = _ref3.end;
  var selection = window.getSelection();

  if (selection) {
    if (node) {
      selection.setBaseAndExtent(node, start, node, end);
    } else if (selection.anchorNode) {
      var anchorNode;

      if (selection.anchorNode.nodeType === Node.TEXT_NODE) {
        anchorNode = selection.anchorNode;
      } else {
        anchorNode = selection.anchorNode.childNodes[0];
      }

      selection.setBaseAndExtent(anchorNode, start, anchorNode, end);
    }
  }
};

var getBeforeCaret = function getBeforeCaret(_ref4, text) {
  var start = _ref4.start;
  return text.slice(0, start);
};

var getAfterCaret = function getAfterCaret(_ref5, text) {
  var end = _ref5.end;
  return text.slice(end);
};

var getLineOf = function getLineOf(_ref, text) {
  var start = _ref.start;

  /*
    1. If there's no \n after it.
      1. If there's \n before it.
        1. Use it to clip it.
      2. Else clip it to the end of the text.
    2. Else
      1. If there's \n before it.
        1. Use it to clip it.
      2. Else clip it to the end of the text.
  */
  var result;
  var isFirstLine = !text.slice(0, start).includes('\n');

  if (isFirstLine) {
    if (text.slice(start).includes('\n')) {
      result = text.slice(0, text.indexOf('\n'));
    } else {
      result = text;
    }
  } else {
    var iA = text.slice(0, start).lastIndexOf('\n');
    var iB = text.indexOf('\n', start);

    if (iB !== -1) {
      result = text.slice(iA + 1, iB);
    } else {
      result = text.slice(iA + 1);
    }
  }

  return result;
};

var getPaddingOf = function getPaddingOf(line) {
  var sum = 0;

  for (var i = 0; i < line.length; i++) {
    if (/\t/.test(line[i])) {
      sum++;
    } else {
      break;
    }
  }

  return '\t'.repeat(sum);
};

var getPadding = function getPadding(text) {
  var caretPosition = getCaretPosition();
  var beforeCaret = getBeforeCaret(caretPosition, text);
  var padding; //? Get the current selected line the (the first one, which will be located above the new line after the new line is injected)

  padding = getPaddingOf(getLineOf(caretPosition, text)); //? Add an additional tab to opening if the character after the caret is an opening bracket except the last one or a mutli line quote

  if (OPENING_BRACKETS.slice(0, 2).includes(beforeCaret[beforeCaret.length - 1]) || beforeCaret[beforeCaret.length - 1] === MUTLI_LINE_QUOTE) padding = padding + '\t';
  return padding;
};

var insert = function insert(text, editor) {
  var caretPosition = getCaretPosition();
  var beforeCaret = getBeforeCaret(caretPosition, editor.innerText);
  var afterCaret = getAfterCaret(caretPosition, editor.innerText);
  var textNode = document.createTextNode("" + beforeCaret + text + afterCaret); //? Remove all of the nodes

  editor.innerHTML = '';
  editor.appendChild(textNode);
};

var handleNewLine = function handleNewLine(editor, wrapper, dynamicHeight, recordHistory) {
  var caretPosition = getCaretPosition(); //? Get the content around the caret

  var beforeCaret = getBeforeCaret(caretPosition, editor.innerText);
  var afterCaret = getAfterCaret(caretPosition, editor.innerText);
  /*
  ? Injecting the updated content
    1. get the padding.
      1. If it exists
        1. If beforeCaret starts in a special character and afterCaret ends with it.
          * A special character is one of these [], {}, `
        2. Else inject it before (beforeCaret).
      2. If it doesn't.
        1. If beforeCaret is empty inject a <br>.
          * To solve the cursor not moving to the new line bug.
        2. If beforeCaret is not empty Inject normally.
  */

  var padding = getPadding(editor.innerText);
  var text;

  if (padding) {
    //? The character after that caret is a quote and the one before it is the same one.
    var isWrappedWithBrackets = OPENING_BRACKETS.includes(beforeCaret[beforeCaret.length - 1]) && afterCaret[0] === CLOSING_BRACKETS[OPENING_BRACKETS.indexOf(beforeCaret[beforeCaret.length - 1])];
    var isWrappedWithMutliLineQuotes = beforeCaret[beforeCaret.length - 1] === MUTLI_LINE_QUOTE && afterCaret[0] === MUTLI_LINE_QUOTE;
    text = "\n" + padding + (isWrappedWithBrackets || isWrappedWithMutliLineQuotes ? "\n" + padding.slice(0, -1) : '');
  } else {
    text = "\n";
  }

  insert(text, editor); //? Move the position of the caret to the next line + the number of tabs

  var nextCaretPosition = beforeCaret.length + 1 + padding.length;
  restoreCaretPosition({
    start: nextCaretPosition,
    end: nextCaretPosition
  });
  recordHistory(editor.innerText, {
    start: nextCaretPosition,
    end: nextCaretPosition
  }); // scrolling in the x-axis

  wrapper.scrollLeft = 0; // scrolling in the y-axis

  if (!dynamicHeight) {
    var lineHeight = parseInt(getComputedStyle(editor).getPropertyValue('line-height'));
    var curLine = getBeforeCaret(getCaretPosition(), editor.innerText).split('').reduce(function (acc, cur) {
      return /\n/g.test(cur) ? acc + 1 : acc;
    }, 0) + 1;
    var paddingTop = parseInt(getComputedStyle(editor).getPropertyValue('padding-top'));

    if (curLine * lineHeight > wrapper.scrollTop + wrapper.clientHeight) {
      wrapper.scrollTop = curLine * lineHeight + paddingTop - wrapper.clientHeight;
    }
  }
};

var handleTab = function handleTab(editor, e) {
  var caretPosition = getCaretPosition();
  var beforeCaret = getBeforeCaret(caretPosition, editor.innerText);
  var afterCaret = getAfterCaret(caretPosition, editor.innerText);

  if (e.shiftKey) {
    if (beforeCaret[beforeCaret.length - 1] === '\t') {
      var textNode = document.createTextNode("" + beforeCaret.slice(0, beforeCaret.length - 1) + afterCaret);
      editor.innerHTML = '';
      editor.appendChild(textNode);
      var nextCursorPosition = beforeCaret.length - 1;
      restoreCaretPosition({
        start: nextCursorPosition,
        end: nextCursorPosition
      });
    }
  } else {
    insert('\t', editor);

    var _nextCursorPosition = beforeCaret.length + 1;

    restoreCaretPosition({
      start: _nextCursorPosition,
      end: _nextCursorPosition
    });
  }
};

var handleSelfClosingCharacters = function handleSelfClosingCharacters(editor, e, recordHistory) {
  /*
    1. If there's no selected text.
          ? Brackets
            1. If it's one of the opening brackets then add a closing one.
        2. If it's one of the closing brackets.
          1. If the cur character is closing then prevent default and skip it.
          2. Else keep it to the browser default (which adds it then skips it).
          ? Single-line quotes
            1. If charater is equal to ' and is wrapped with "" don't prevent adding two quotes.
        2. Else
          1. count down the number of the same character in the line.
            1. If it's a single quote then don't count the wrapped quotes.
              * (1) check this in code to see how it can be implemented.
            2. Else count any appearance of the character.
          2. If the number is even and character before the caret isn't equal to it.
            1. Add two quotes.
          3. If the number is odd add one.
            3. Move the caret one step forward.
          ? Mutli-line quote
            1. If there's an even number of the caracter in the whole text and the character before it the caret isn't equal to it.
          - Add a new character and it's closing one.
        - If there's an odd number.
        2. Move the caret one step forward.
      */
  var selection = window.getSelection();
  var caretPosition = getCaretPosition();
  var character = e.key;
  var beforeCaret = getBeforeCaret(caretPosition, editor.innerText);
  var afterCaret = getAfterCaret(caretPosition, editor.innerText);

  if (selection.type === 'Caret') {
    var line = getLineOf(caretPosition, editor.innerText); // Brackets

    if (OPENING_BRACKETS.includes(character)) {
      e.preventDefault();
      insert("" + character + CLOSING_BRACKETS[OPENING_BRACKETS.indexOf(character)], editor);
      var nextCaretPosition = beforeCaret.length + 1;
      restoreCaretPosition({
        start: nextCaretPosition,
        end: nextCaretPosition
      });
      recordHistory(editor.innerText, {
        start: nextCaretPosition,
        end: nextCaretPosition
      });
    }

    if (CLOSING_BRACKETS.includes(character)) {
      if (afterCaret[0] === character) {
        e.preventDefault();

        var _nextCaretPosition = beforeCaret.length + 1;

        restoreCaretPosition({
          start: _nextCaretPosition,
          end: _nextCaretPosition
        });
      }
    } // Single line quotes


    if (SINGLE_LINE_QUOTES.includes(character)) {
      e.preventDefault(); // I know, not the best name 😑

      var isCharactersInLineEven = line.split('').reduce(function (acc, cur) {
        return cur === character ? acc + 1 : acc;
      }, 0) % 2 === 0;

      var _nextCaretPosition2 = beforeCaret.length + 1;

      if (isCharactersInLineEven) {
        if (afterCaret[0] !== character) {
          insert(character.repeat(2), editor);
          restoreCaretPosition({
            start: _nextCaretPosition2,
            end: _nextCaretPosition2
          });
          recordHistory(editor.innerText, {
            start: _nextCaretPosition2,
            end: _nextCaretPosition2
          });
        } else {
          restoreCaretPosition({
            start: _nextCaretPosition2,
            end: _nextCaretPosition2
          });
        }
      } else {
        insert(character, editor);
        restoreCaretPosition({
          start: _nextCaretPosition2,
          end: _nextCaretPosition2
        });
      }
    } // Mutli line quote


    if (character === MUTLI_LINE_QUOTE) {
      e.preventDefault(); // I also know, not the best name 😶

      var isCharactersInWholeTextEven = editor.innerText.split('').reduce(function (acc, cur) {
        return cur === MUTLI_LINE_QUOTE ? acc + 1 : acc;
      }, 0) % 2 === 0;

      var _nextCaretPosition3 = beforeCaret.length + 1;

      if (isCharactersInWholeTextEven) {
        if (beforeCaret[0] !== character) {
          insert(character.repeat(2), editor);
          restoreCaretPosition({
            start: _nextCaretPosition3,
            end: _nextCaretPosition3
          });
          recordHistory(editor.innerText, {
            start: _nextCaretPosition3,
            end: _nextCaretPosition3
          });
        } else {
          restoreCaretPosition({
            start: _nextCaretPosition3,
            end: _nextCaretPosition3
          });
        }
      } else {
        insert(character, editor);
        editor.innerHTML = "" + beforeCaret + character + afterCaret;
        restoreCaretPosition({
          start: _nextCaretPosition3,
          end: _nextCaretPosition3
        });
      }
    }
  }

  if (selection.type === 'Range') {
    var selectedText = editor.innerText.slice(beforeCaret.length, editor.innerText.length - afterCaret.length);
    var closingCharacter;
    var isBracket = OPENING_BRACKETS.includes(character);
    var isQuote = [].concat(SINGLE_LINE_QUOTES, [MUTLI_LINE_QUOTE]).includes(character);

    if (isBracket) {
      closingCharacter = CLOSING_BRACKETS[OPENING_BRACKETS.indexOf(character)];
    }

    if (isQuote) {
      closingCharacter = character;
    }

    if (isBracket || isQuote) {
      e.preventDefault(); //? Wrap the selected text

      insert("" + character + selectedText + closingCharacter, editor); //? Re select the previously selected text

      var _nextCaretPosition4 = {
        start: beforeCaret.length + 1,
        end: beforeCaret.length + selectedText.length + 1
      };
      restoreCaretPosition(_nextCaretPosition4);
      recordHistory(editor.innerText, _nextCaretPosition4);
    }
  }
};

var handleSpecialCharactersRemoving = function handleSpecialCharactersRemoving(editor, e) {
  var caretPosition = getCaretPosition();
  var beforeCaret = getBeforeCaret(caretPosition, editor.innerText);
  var afterCaret = getAfterCaret(caretPosition, editor.innerText); //? The character after the caret is an opening bracket and the character before it is its closing one

  var isWrappedWithBrackets = OPENING_BRACKETS.includes(beforeCaret[beforeCaret.length - 1]) && afterCaret[0] === CLOSING_BRACKETS[OPENING_BRACKETS.indexOf(beforeCaret[beforeCaret.length - 1])]; //? The character after that caret is a quote and the one before it is the same one.

  var isWrappedWithQuotes = [].concat(SINGLE_LINE_QUOTES, [MUTLI_LINE_QUOTE]).includes(beforeCaret[beforeCaret.length - 1]) && afterCaret[0] === beforeCaret[beforeCaret.length - 1];

  if (isWrappedWithBrackets || isWrappedWithQuotes) {
    e.preventDefault();
    var textNode = document.createTextNode("" + beforeCaret.slice(0, -1) + afterCaret.slice(1));
    editor.innerHTML = '';
    editor.appendChild(textNode);
    var nextCaretPosition = beforeCaret.length - 1;
    restoreCaretPosition({
      start: nextCaretPosition,
      end: nextCaretPosition
    });
  }
};

var reducer = function reducer(state, action) {
  if (action.type === 0
  /* RECORD */
  ) {
    /*
      Recording history
        1. Push `present` to `past`.
        2. Set `present` to `action.payload`.
        3. Clear `future`.
    */
    return {
      past: [].concat(state.past, [state.present]),
      present: action.payload,
      future: []
    };
  }

  if (action.type === 1
  /* UNDO */
  ) {
    /*
      Undoing history
        1. If `past` isn't empty.
          1. Pop `past`.
          2. Set `present` to the popped value.
          3. Insert `present` to `future`.
    */
    if (state.past.length) {
      return {
        past: state.past.slice(0, -1),
        present: state.past[state.past.length - 1],
        future: [state.present].concat(state.future)
      };
    }
  }

  if (action.type === 2
  /* REDO */
  ) {
    /*
      Redoing history
        1. If future isen't empty
          1. Push `present` to `past`.
          2. Set `present` to the first item of `future`.
          3. Remove the first item of `future`.
    */
    if (state.future.length) {
      return {
        past: [].concat(state.past, [state.present]),
        present: state.future[0],
        future: state.future.slice(1)
      };
    }
  }

  return state;
};

var initialState = {
  past: [],
  present: {
    text: '',
    position: {
      start: 0,
      end: 0
    }
  },
  future: []
};

var handleCharacter = function handleCharacter(editor, character, recordHistory) {
  var caretPosition = getCaretPosition();
  var beforeCaret = getBeforeCaret(caretPosition, editor.innerHTML);
  insert(character, editor);
  var nextCaretPosition = beforeCaret.length + 1;
  restoreCaretPosition({
    start: nextCaretPosition,
    end: nextCaretPosition
  });
  recordHistory(editor.innerText, {
    start: nextCaretPosition,
    end: nextCaretPosition
  });
};

var handleRangeRemoving = function handleRangeRemoving(editor, recordHistory) {
  var caretPosition = getCaretPosition();
  var beforeCaret = getBeforeCaret(caretPosition, editor.innerText);
  recordHistory(editor.innerText, caretPosition);
  insert('', editor);
  var nextCaretPosition = beforeCaret.length;
  restoreCaretPosition({
    start: nextCaretPosition,
    end: nextCaretPosition
  });
};

function styleInject(css, ref) {
  if ( ref === void 0 ) ref = {};
  var insertAt = ref.insertAt;

  if (!css || typeof document === 'undefined') { return; }

  var head = document.head || document.getElementsByTagName('head')[0];
  var style = document.createElement('style');
  style.type = 'text/css';

  if (insertAt === 'top') {
    if (head.firstChild) {
      head.insertBefore(style, head.firstChild);
    } else {
      head.appendChild(style);
    }
  } else {
    head.appendChild(style);
  }

  if (style.styleSheet) {
    style.styleSheet.cssText = css;
  } else {
    style.appendChild(document.createTextNode(css));
  }
}

var css_248z = "@charset \"UTF-8\";\n@import url(\"https://fonts.googleapis.com/css2?family=Azeret+Mono:ital,wght@0,400;0,500;1,500&display=swap\");\n.editor {\n  display: flex;\n  width: 100%;\n  height: 100%;\n  box-sizing: border-box;\n  font-family: \"Azeret Mono\", monospace;\n  font-size: 0.8rem;\n  line-height: 1.25rem;\n}\n.editor--light {\n  --color-fg: #383a42;\n  --color-fg-weak: #383a42aa;\n  --color-bg: #fafafa;\n}\n.editor--dark {\n  --color-fg: #abb2bf;\n  --color-fg-weak: #abb2bfaa;\n  --color-bg: #282c34;\n}\n.editor *, .editor *::after, .editor *::before {\n  box-sizing: inherit;\n}\n.editor__lines-numbers {\n  flex: 0 0 max-content;\n  display: flex;\n  flex-direction: column;\n  align-items: center;\n  padding: 0.5rem;\n  background-color: var(--color-bg);\n  border-right: 1px solid var(--color-fg);\n  min-height: 100%;\n  height: max-content;\n  color: var(--color-fg-weak);\n  font-weight: 400;\n}\n.editor--dynamic .editor__line-numbers {\n  justify-content: space-around;\n  line-height: 0;\n}\n.editor__main {\n  position: relative;\n  flex: 1;\n  font-weight: 500;\n  tab-size: 4;\n}\n.editor--dymanic .editor__main {\n  line-height: 1.75;\n}\n.editor__textarea, .editor__previewer {\n  padding: 0.5rem;\n  outline: none;\n  min-width: 100%;\n  min-height: 100%;\n  width: max-content;\n  height: max-content;\n}\n.editor__textarea::after, .editor__previewer::after {\n  content: \"​\";\n}\n.editor__textarea::-webkit-scrollbar, .editor__previewer::-webkit-scrollbar {\n  height: 0;\n}\n.editor__textarea {\n  position: relative;\n  z-index: 10;\n  caret-color: var(--color-fg);\n  color: #00000005;\n}\n.editor__previewer {\n  position: absolute;\n  top: 0;\n  color: var(--color-fg);\n  background-color: var(--color-bg);\n}\n.editor--light .editor__previewer .comment,\n.editor--light .editor__previewer .quote {\n  color: #a0a1a7;\n}\n.editor--light .editor__previewer .doctag,\n.editor--light .editor__previewer .keyword,\n.editor--light .editor__previewer .formula {\n  color: #a626a4;\n}\n.editor--light .editor__previewer .section,\n.editor--light .editor__previewer .name,\n.editor--light .editor__previewer .selector-tag,\n.editor--light .editor__previewer .deletion,\n.editor--light .editor__previewer .subst {\n  color: #e45649;\n}\n.editor--light .editor__previewer .literal {\n  color: #0184bb;\n}\n.editor--light .editor__previewer .string,\n.editor--light .editor__previewer .regexp,\n.editor--light .editor__previewer .addition,\n.editor--light .editor__previewer .attribute,\n.editor--light .editor__previewer .meta .string {\n  color: #50a14f;\n}\n.editor--light .editor__previewer .attr,\n.editor--light .editor__previewer .variable,\n.editor--light .editor__previewer .template-variable,\n.editor--light .editor__previewer .type,\n.editor--light .editor__previewer .selector-class,\n.editor--light .editor__previewer .selector-attr,\n.editor--light .editor__previewer .selector-pseudo,\n.editor--light .editor__previewer .number {\n  color: #986801;\n}\n.editor--light .editor__previewer .symbol,\n.editor--light .editor__previewer .bullet,\n.editor--light .editor__previewer .link,\n.editor--light .editor__previewer .meta,\n.editor--light .editor__previewer .selector-id,\n.editor--light .editor__previewer .title {\n  color: #4078f2;\n}\n.editor--light .editor__previewer .built_in,\n.editor--light .editor__previewer .title.class_,\n.editor--light .editor__previewer .class .title {\n  color: #c18401;\n}\n.editor--dark .editor__previewer .comment,\n.editor--dark .editor__previewer .quote {\n  color: #5c6370;\n}\n.editor--dark .editor__previewer .doctag,\n.editor--dark .editor__previewer .keyword,\n.editor--dark .editor__previewer .formula {\n  color: #c678dd;\n}\n.editor--dark .editor__previewer .section,\n.editor--dark .editor__previewer .name,\n.editor--dark .editor__previewer .selector-tag,\n.editor--dark .editor__previewer .deletion,\n.editor--dark .editor__previewer .subst {\n  color: #e06c75;\n}\n.editor--dark .editor__previewer .literal {\n  color: #56b6c2;\n}\n.editor--dark .editor__previewer .string,\n.editor--dark .editor__previewer .regexp,\n.editor--dark .editor__previewer .addition,\n.editor--dark .editor__previewer .attribute,\n.editor--dark .editor__previewer .meta .string {\n  color: #98c379;\n}\n.editor--dark .editor__previewer .attr,\n.editor--dark .editor__previewer .variable,\n.editor--dark .editor__previewer .template-variable,\n.editor--dark .editor__previewer .type,\n.editor--dark .editor__previewer .selector-class,\n.editor--dark .editor__previewer .selector-attr,\n.editor--dark .editor__previewer .selector-pseudo,\n.editor--dark .editor__previewer .number {\n  color: #d19a66;\n}\n.editor--dark .editor__previewer .symbol,\n.editor--dark .editor__previewer .bullet,\n.editor--dark .editor__previewer .link,\n.editor--dark .editor__previewer .meta,\n.editor--dark .editor__previewer .selector-id,\n.editor--dark .editor__previewer .title {\n  color: #61aeee;\n}\n.editor--dark .editor__previewer .built_in,\n.editor--dark .editor__previewer .title.class_,\n.editor--dark .editor__previewer .class .title {\n  color: #e6c07b;\n}";
styleInject(css_248z);

var handlePaste = function handlePaste(e, editor, recordHistory) {
  /*
      ? Handling copies
          1. Record the history.
          2. Insert the new text as plain text.
          3. Record the history.
  */
  e.preventDefault();
  var caretPosition = getCaretPosition();
  var beforeCaret = getBeforeCaret(caretPosition, editor.innerText);
  var copiedText = e.clipboardData.getData('text/plain');
  recordHistory(editor.innerText, caretPosition);
  insert(copiedText, editor);
  var newCaretPosition = beforeCaret.length + copiedText.length;
  restoreCaretPosition({
    start: newCaretPosition,
    end: newCaretPosition
  });
  recordHistory(editor.innerText, {
    start: newCaretPosition,
    end: newCaretPosition
  });
};

var CodeEditor = function CodeEditor(_ref) {
  var language = _ref.language,
      value = _ref.value,
      onChange = _ref.onChange,
      _ref$theme = _ref.theme,
      theme = _ref$theme === void 0 ? 'light' : _ref$theme,
      _ref$height = _ref.height,
      height = _ref$height === void 0 ? 'auto' : _ref$height,
      _ref$spellCheck = _ref.spellCheck,
      spellCheck = _ref$spellCheck === void 0 ? false : _ref$spellCheck,
      _ref$handleHistory = _ref.handleHistory,
      handleHistory = _ref$handleHistory === void 0 ? true : _ref$handleHistory,
      _ref$handleSpecialCha = _ref.handleSpecialCharacters,
      handleSpecialCharacters = _ref$handleSpecialCha === void 0 ? true : _ref$handleSpecialCha,
      _ref$highlight = _ref.highlight,
      highlight = _ref$highlight === void 0 ? true : _ref$highlight,
      _ref$lineNumbers = _ref.lineNumbers,
      lineNumbers = _ref$lineNumbers === void 0 ? true : _ref$lineNumbers,
      className = _ref.className;

  var _useReducer = useReducer(reducer, initialState),
      state = _useReducer[0],
      send = _useReducer[1];

  var wrapperRef = useRef(null);
  var editorRef = useRef(null);
  var previewerRef = useRef(null);

  var _useState = useState(0),
      lineNumber = _useState[0],
      setLineNumber = _useState[1];

  var _useState2 = useState(''),
      curText = _useState2[0],
      setCurText = _useState2[1];

  var isLanguageSupported = lowlight.listLanguages().includes(language);

  var calculateLineNumber = function calculateLineNumber(text) {
    setLineNumber(text.split('').reduce(function (acc, cur) {
      return /\n/g.test(cur) ? acc + 1 : acc;
    }, 0) + 1);
  };

  useEffect(function () {
    if (!isLanguageSupported) {
      //TODO add a ref to the doc when it's written
      console.error('💥 The language that you added is not supported by default\nYou can add it by importing the language, then regestering it with the command `hljs.registerLanguage("javascript", javascript).');
    }
  }, []);
  useEffect(function () {
    calculateLineNumber(curText);
  }, [curText]);
  useEffect(function () {
    // highlighting logic
    var previewer = previewerRef.current;

    if (highlight && isLanguageSupported) {
      var tree = lowlight.highlight(language, curText, {
        prefix: ''
      });
      previewer.innerHTML = toHtml(tree);
    } else {
      previewer.innerText = curText;
    }

    onChange(curText);
  }, [curText, highlight]); // override the input whenever it's changed by the reducer

  useEffect(function () {
    var editor = editorRef.current;

    if (state.present.text !== editor.innerText) {
      var textNode = document.createTextNode(state.present.text);
      editor.innerHTML = '';
      editor.appendChild(textNode);
      restoreCaretPosition(state.present.position, textNode); //? Whenever we undo or redo

      setCurText(editor.innerText);
    }
  }, [state.present]); // override the editor if value !== curText

  useEffect(function () {
    if (value !== curText) {
      recordHistory(value, {
        start: value.length,
        end: value.length
      });
    }
  }, [value]);
  /*
      ? When is the handleHistory recorded
          1. Whenever a space or dot is added.
          2. When a new line is injected.
          3. When double characters are added.
          4. Undoing or redoing and the present isn't equal to the cur input value.
          5. If some text is selected and we clicked a button that will delete it.
  */

  var recordHistory = function recordHistory(text, position) {
    send({
      type: 0
      /* RECORD */
      ,
      payload: {
        text: text,
        position: position
      }
    });
  };

  var keyDownHandler = function keyDownHandler(e) {
    var editor = editorRef.current;
    var warpper = wrapperRef.current;
    var textAfter = editor.innerText;
    var caretPosition = getCaretPosition();
    var isUndo = e.nativeEvent.code === 'KeyZ' && e.ctrlKey;
    var isRedo = e.nativeEvent.code === 'KeyY' && e.ctrlKey;
    var isSelectAll = e.nativeEvent.code === 'KeyA' && e.ctrlKey;
    var isCopy = e.nativeEvent.code === 'KeyC' && e.ctrlKey;
    var isCut = e.nativeEvent.code === 'KeyX' && e.ctrlKey;
    var isPaste = e.nativeEvent.code === 'KeyV' && e.ctrlKey;

    if (isUndo || isRedo || isSelectAll || e.key === 'Enter' || e.key === 'Tab' || e.nativeEvent.code === 'Space' || e.key === '.') {
      e.preventDefault();
    } // 5.


    if (caretPosition.start !== caretPosition.end && ![].concat(OPENING_BRACKETS, SINGLE_LINE_QUOTES, [MUTLI_LINE_QUOTE]).includes(e.key) && !isUndo && !isRedo && !isSelectAll && !isCopy && !isPaste && !isCut && e.key !== 'Meta' && e.key !== 'Control' && e.key !== 'Alt' && e.key !== 'Tab' && e.nativeEvent.code !== 'Space' && e.key !== 'Enter' && !e.nativeEvent.code.startsWith('Arrow')) {
      e.preventDefault();

      if (e.key === 'Backspace') {
        handleRangeRemoving(editor, recordHistory);
        recordHistory(editor.innerText, getCaretPosition());
      } else {
        recordHistory(editor.innerText, caretPosition);
        handleCharacter(editor, e.key, recordHistory);
      }
    }

    if (handleHistory) {
      if (isUndo) {
        // 4.
        if (editor.innerText !== state.present.text) {
          recordHistory(editor.innerText, caretPosition);
        } //* To fix updating issue


        setTimeout(function () {
          return send({
            type: 1
            /* UNDO */

          });
        });
      }

      if (isRedo) {
        // 4.
        if (editor.innerText !== state.present.text) {
          recordHistory(editor.innerText, caretPosition);
        } else {
          send({
            type: 2
            /* REDO */

          });
        }
      }
    }

    if (isSelectAll) {
      restoreCaretPosition({
        start: 0,
        end: editor.innerText.length
      });
    }

    if (e.key === 'Enter') {
      // 2.
      handleNewLine(editor, warpper, height === 'auto', recordHistory);
    }

    if (e.key === 'Tab') {
      handleTab(editor, e);
    }

    if (SPECIAL_CHARACTERS.includes(e.key)) {
      // 3.
      if (handleSpecialCharacters) {
        handleSelfClosingCharacters(editor, e, recordHistory);
      }
    }

    if (e.key === 'Backspace') {
      if (handleSpecialCharacters) {
        handleSpecialCharactersRemoving(editor, e);
      } //* Removing the default behavior of adding a <br> if the editor is empty


      if (!editor.innerHTML) {
        e.preventDefault();
      }
    }

    if (e.nativeEvent.code === 'Space') {
      // 1.
      handleCharacter(editor, ' ', recordHistory);
    }

    if (e.key === '.') {
      // 1.
      handleCharacter(editor, '.', recordHistory);
    } //? Whenever something is changed by the handlers


    var textBefore = editor.innerText;

    if (textAfter !== textBefore) {
      setCurText(textBefore);
    }
    /*
        ? Fixing going to the last line with caret bug
            1. If the down arrow is pressed and we're after the empty last empty line.
            2. If the right arrow is pressed and we're in the last character after the last empty line.
    */


    if (textBefore.endsWith('\n')) {
      var _caretPosition = getCaretPosition();

      if (e.nativeEvent.code === 'ArrowDown' && _caretPosition.start === _caretPosition.end && _caretPosition.start <= textBefore.lastIndexOf('\n') && _caretPosition.start > textBefore.slice(0, -1).lastIndexOf('\n')) {
        e.preventDefault();
        restoreCaretPosition({
          start: textBefore.length,
          end: textBefore.length
        });
      }

      if (e.nativeEvent.code === 'ArrowRight' && _caretPosition.start === textBefore.lastIndexOf('\n')) {
        e.preventDefault();
        restoreCaretPosition({
          start: textBefore.length,
          end: textBefore.length
        });
      }
    }
  };

  return React.createElement("div", {
    style: {
      height: height,
      width: '100%',
      overflowY: height === 'auto' ? 'hidden' : 'scroll',
      overflowX: 'scroll'
    },
    className: className,
    ref: wrapperRef
  }, React.createElement("div", {
    className: "editor editor--" + theme + " " + (height === 'auto' ? 'editor--dynamic' : '')
  }, lineNumbers && React.createElement("div", {
    className: "editor__lines-numbers"
  }, Array.from({
    length: lineNumber
  }).map(function (_, i) {
    return React.createElement("span", {
      key: i
    }, i + 1);
  })), React.createElement("div", {
    className: "editor__main"
  }, React.createElement("div", {
    className: "editor__textarea",
    style: {
      whiteSpace: 'pre'
    },
    ref: editorRef,
    contentEditable: true,
    spellCheck: spellCheck,
    onInput: function onInput() {
      //? Whenever a new line is removed by the default behavior of the browser
      var editor = editorRef.current;
      setCurText(editor.innerText);
    },
    onKeyDown: keyDownHandler,
    onPaste: function onPaste(e) {
      var editor = editorRef.current;
      handlePaste(e, editor, recordHistory);
      setCurText(editor.innerText);
    }
  }), React.createElement("div", {
    className: "editor__previewer",
    ref: previewerRef,
    style: {
      whiteSpace: 'pre'
    }
  }))));
};

export default CodeEditor;
//# sourceMappingURL=react-throwcode.esm.js.map
